/* $Id: create_table_4.c,v 1.5 2010/09/09 12:09:57 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2005 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2010/09/09 12:09:57 $
 * $Revision: 1.5 $
 * $Name: vcam-1_3_0 $
 */

#include <stdio.h>
#include <math.h>
#include "imcore.h"
#include "util.h"
#include "../vircam_fits.h"

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Initialise type 4 catalogue (object mask)
  
    \par Name:
        tabinit_4
    \par Purpose:
        Initialise type 4 catalogue (object mask)
    \par Description:
        Type 4 catalogue is initialised. This is in fact an object mask.
    \par Language:
        C
    \param ap
        The current ap structure
    \returns
        Nothing
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void tabinit_4(ap_t *ap) {
    long npts,i;
    cpl_binary *opm;

    /* The output table doesn't exist */

    tab = NULL;
    npts = (ap->lsiz)*(ap->csiz);
    ap->opmask = cpl_mask_new(ap->lsiz,ap->csiz);
    opm = cpl_mask_get_data(ap->opmask);
    for (i = 0; i < npts; i++) 
	opm[i] = 0;

    /* Undefined RA and Dec columns */

    imcore_xcol = -1;
    imcore_ycol = -1;
    
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Dummy seeing routine for type 4 catalogue (object mask)
  
    \par Name:
        do_seeing_4
    \par Purpose:
        Do seeing estimate for type 4 catalogue (object mask)
    \par Description:
        This is a dummy routine
    \par Language:
        C
    \param ap
        The current ap structure
    \retval VIR_OK
        If all is well. Currently this is the only return value
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int do_seeing_4(ap_t *ap) {

    /* Get out of here */

    ap->fwhm = 0.0;
    return(VIR_OK);
}
        
/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Process results for type 4 catalogue (object mask)
  
    \par Name:
        process_results_4
    \par Purpose:
        Create the results for objects in a type 4 catalogue (object mask)
    \par Description:
        The positions for pixels identified as being part of a detected
	object are flagged in the object pixel mask.
    \par Language:
        C
    \param ap
        The current ap structure
    \retval VIR_OK
        If all is well. This is currently the only value
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/
        
extern int process_results_4(ap_t *ap) {
    int i,j,np;
    long nx;
    plstruct *plarray;
    cpl_binary *opm;

    /* Loop for each object in the array */

    opm = cpl_mask_get_data(ap->opmask);
    plarray = ap->plarray;
    np = ap->npl_pix;
    nx = ap->lsiz;
    for (i = 0; i < np; i++) {
	j = nx*(plarray[i].y - 1) + plarray[i].x - 1;
	opm[j] = 1;
    }

    /* Get outta here */

    return(VIR_OK);
}

        
/*---------------------------------------------------------------------------*/
/**
    \ingroup cataloguemodules
    \brief Close object mask
  
    \par Name:
        tabclose_4
    \par Purpose:
        Close the object mask
    \par Description:
        The object mask array in the ap structure is freed and the internal
	mask in the input image is marked with the object locations.
    \par Language:
        C
    \param ap
        The current ap structure
    \retval VIR_OK
        If all is well. This is currently the only value
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int tabclose_4(ap_t *ap) {

    cpl_image_reject_from_mask(ap->inframe,ap->opmask);
    cpl_mask_delete(ap->opmask);
    return(VIR_OK);
}

/**@}*/

/*

$Log: create_table_4.c,v $
Revision 1.5  2010/09/09 12:09:57  jim
Added docs

Revision 1.4  2007/05/03 11:15:34  jim
Fixed little problem with table wcs

Revision 1.3  2007/05/02 09:11:35  jim
Modified to allow for inclusion of table WCS keywords into FITS header

Revision 1.2  2006/03/01 10:31:29  jim
Now uses new vir_fits objects

Revision 1.1  2005/09/13 13:25:28  jim
Initial entry after modifications to make cpl compliant


*/
