/* $Id: vircam_imstack.c,v 1.15 2012/01/15 17:40:09 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2005 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2012/01/15 17:40:09 $
 * $Revision: 1.15 $
 * $Name: vcam-1_3_0 $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <cpl.h>
#include "vircam_mods.h"
#include "vircam_utils.h"
#include "vircam_fits.h"
#include "vircam_stats.h"
#include "vircam_pfits.h"
#include "vircam_wcsutils.h"
#include "vircam_tfits.h"

typedef struct {
    double     crval[2];
    double     crpix[2];
    double     cd[4];
    double     secd;
    double     tand;
    double     pv21;
    double     pv23;
    double     pv25;
    long       nx;
    long       ny;
} mywcs;

typedef struct {
    int        nx;
    int        ny;
    vir_fits   *fname;
    vir_fits   *conf;
    float      *data;
    int        *cdata;
    cpl_binary *bpm;
    mywcs      *vwcs;
    float      xoff;
    float      yoff;
    cpl_array  *trans;
    float      sky;
    float      skydiff;
    float      noise;
    float      expscale;
    float      weight;
} dstrct;

static dstrct *fileptrs = NULL;
static vir_fits *outfits = NULL;
static vir_fits *outconf = NULL;
static mywcs *outwcs = NULL;

static float lsig;
static float hsig;
static int   nxo;
static int   nyo;
static float *outdata;
static int   *outdatac;
static int   nim;
static float sumweight;

static void outloc(mywcs *win, double xin, double yin, mywcs *wout, 
		   cpl_array *trans, double *xout, double *yout);
static void backgrnd_ov(void);
static void skyest(float *data, int *cdata, long npts, float thresh, 
		   float *skymed, float *skynoise);
static cpl_array *transinit(void);
static void fileptrs_close(void);
static cpl_table *mknewtab(vir_tfits *catref, dstrct *dd);
static void output_files(void);
static int stack_nn(void);
static int stack_lin(void);
static void do_averages(int ncontrib, float *data, float *wconf, float *conf,
			unsigned char *id, float lclip, float hclip, 
			float extra, short int *lowcl, short int *highcl, 
			float *outval, float *outvalc);
static mywcs *getwcsinfo(cpl_propertylist *plist);
static void diffxy(mywcs *wref, mywcs *wprog, float *dx, float *dy);
static void tidy(void);

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Stack images into a mean image using WCS info 

    \par Name:
        vircam_imstack
    \par Purpose:
        Stack images into a mean image using WCS info
    \par Description:
        A list of frames is given. A mean image is stack by using the 
	given WCS to work out the shift of an input pixel relative to the
	grid defined by the first image in the list. At first all are
	stacked using a nearest neighbour algorithm. Any discordant pixels
	are marked at this stage. If a different stacking algorithm is 
	required, then the images are restacked, but with the discordant
	pixels found earlier removed. If object catalogues for each input
	image are available, then these are used to further refine the 
	coordinate transformation.
    \par Language:
        C
    \param inf
        Input image list
    \param inconf
        Input list of confidence maps
    \param cats
        Input list of object catalogues. If NULL, then no refinement will
	be done.
    \param nimages
        The number of input images and (if given) the number of input 
	object catalogues
    \param nconfs
        The number of input confidence maps
    \param lthr
        The lower rejection threshold
    \param hthr
        The upper rejection threshold
    \param method
        The interpolation method:
	- 0: Nearest neighbour
        - 1: Bi-linear interpolation
    \param out
        The output stacked image
    \param outc
        The output confidence map
    \param status
        The usual input/output status variable
    \retval VIR_OK 
        if everything is ok
    \retval VIR_FATAL 
        if there are no images to combine or there is a mismatch between
	the confidence map sizes and image sizes.
    \par DRS headers:
        The following DRS keywords are written to the drs propertylist
	- \b PROV****
	    The provenance keywords
    \author
        Jim Lewis, CASU

 */
/*---------------------------------------------------------------------------*/

extern int vircam_imstack(vir_fits **inf, vir_fits **inconf, vir_tfits **cats,
			  int nimages, int nconfs, float lthr, float hthr,
			  int method, vir_fits **out, vir_fits **outc, 
			  int *status) {

    int i,nm;
    dstrct *dd;
    float expref,exposure,xoff,yoff;
    cpl_matrix *in;
    vir_tfits *catref;
    cpl_table *newtab,*mtab;
    const char *fctid = "vircam_imstack";

    /* Inherited status */

    *out = NULL;
    *outc = NULL;
    if (*status != VIR_OK)
        return(*status);

    /* Is there any point in being here? */

    if (nimages == 0) {
        cpl_msg_error(fctid,"No input files to combine");
        tidy();
        FATAL_ERROR
    }

    /* Initialise some global variables */

    lsig = lthr;
    hsig = hthr;
    nim = nimages;

    /* Initialise a useful matrix */

    in = cpl_matrix_new(1,2);
    cpl_matrix_set(in,0,0,0.0);
    cpl_matrix_set(in,0,1,0.0);

    /* Allocate file struct array and fill in some values */

    fileptrs = cpl_malloc(nimages*sizeof(dstrct));
    (void)vircam_pfits_get_exptime(vircam_fits_get_phu(inf[0]),&exposure);
    expref = max(0.5,exposure);
    catref = NULL;
    for (i = 0; i < nimages; i++) {
	dd = fileptrs + i;
        dd->fname = inf[i];
	dd->data = cpl_image_get_data_float(vircam_fits_get_image(inf[i]));
	if (nconfs == 0) {
	    dd->conf = NULL;
	} else if (nconfs == 1) {
	    dd->conf = inconf[0];
  	    dd->cdata = cpl_image_get_data_int(vircam_fits_get_image(inconf[0]));
	} else {
	    dd->conf = inconf[i];
  	    dd->cdata = cpl_image_get_data_int(vircam_fits_get_image(inconf[i]));
	}
	(void)vircam_pfits_get_exptime(vircam_fits_get_phu(dd->fname),&exposure);
	dd->expscale = exposure/expref;

	/* Zero the bad pixel mask */

	cpl_image_accept_all(vircam_fits_get_image(inf[i]));
	dd->bpm = cpl_mask_get_data(cpl_image_get_bpm(vircam_fits_get_image(inf[i])));

	/* Read the WCS from the header */

	dd->vwcs = getwcsinfo(vircam_fits_get_ehu(inf[i]));

        /* Get the image size */

        dd->nx = (int)cpl_image_get_size_x(vircam_fits_get_image(dd->fname));
        dd->ny = (int)cpl_image_get_size_y(vircam_fits_get_image(dd->fname));
	
        /* Double check to make sure the confidence maps and images have the
	   same dimensions */

	if ((int)cpl_image_get_size_x(vircam_fits_get_image(dd->conf)) != dd->nx || 
	    (int)cpl_image_get_size_y(vircam_fits_get_image(dd->conf)) != dd->ny) {
	    cpl_msg_error(fctid,"Image %s and Confidence map %s don't match",
			  vircam_fits_get_fullname(dd->fname),
			  vircam_fits_get_fullname(dd->conf));
	    cpl_matrix_delete(in);
	    tidy();
	    FATAL_ERROR
	}

	/* Get a rough shift between the frames and the first one */

	diffxy(fileptrs->vwcs,dd->vwcs,&(dd->xoff),&(dd->yoff));

        /* Do we have catalogues? If so, then work out the transformation 
           corrections from them */

        if (cats != NULL) {
	    if (catref == NULL && cats[i] != NULL)
		catref = cats[i];
	    if (i == 0 || cats[i] == NULL) {
		dd->trans = transinit();
	    } else {
		newtab = mknewtab(catref,dd);		
   	        (void)vircam_matchxy(vircam_tfits_get_table(cats[i]),newtab,
				     1024,&xoff,&yoff,&nm,&mtab,status);
		if (*status != VIR_OK) {
		    freearray(dd->trans);
		    dd->trans = transinit();
		    *status = VIR_OK;
		} else { 
   		    (void)vircam_platexy(mtab,6,&(dd->trans),status);
                }
		freetable(mtab);
		freetable(newtab);
	    }
	} else {
	    dd->trans = NULL;
	}
    }
    cpl_matrix_delete(in);

    /* Get the background levels in the overlap regions */

    backgrnd_ov();

    /* Create the output file */

    output_files();

    /* Ok time to do the stacking. Our first job is to do a nearest neighbour
       stack in order to find the objects that will be rejected. If we 
       only want the NN algorithm, then we're done. Otherwise we can use the
       rejection information to restack using any other algorithm */

    stack_nn();

    /* Switch for final stacking method */

    switch (method) {
    case 0:
	break;
    case 1:
	stack_lin();
	break;
    default:
	break;
    }

    /* Add the provenance cards to the header */

    vircam_prov(vircam_fits_get_ehu(outfits),inf,nimages);

    /* Assign some output info, tidy and get out of here */

    *out = outfits;
    *outc = outconf;
    tidy();
    return(VIR_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        stack_lin
    \par Purpose:
        Do a stack with bi-linear interpolation
    \par Description:
        Do a stack with bi-linear interpolation. No rejection is done 
	here as it is assumed that this has been done with the nearest
	neighbour algorithm and any bad pixels have been flagged in the
	image masks
    \par Language:
        C
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int stack_lin(void) {
    cpl_image *outim,*outimc;
    double oxf,oyf;
    int npts,i,*cdata,ix,iy,ind1,ind,ox,oy,m,jj,ii,nn,nx,ny,n;
    float *workbufw,*data,wt,dx,dy,w[4],value,cval;
    float med,noise;
    dstrct *dd;

    /* Output file info */

    outim = vircam_fits_get_image(outfits);
    outimc = vircam_fits_get_image(outconf);
    nxo = (int)cpl_image_get_size_x(outim);
    nyo = (int)cpl_image_get_size_y(outim);
    npts = nxo*nyo;

    /* Get workspace for collecting info over the whole output map */

    workbufw = cpl_calloc(nxo*nyo,sizeof(*workbufw));

    /* Get output buffers */

    outdata = cpl_image_get_data_float(outim);
    outdatac = cpl_image_get_data_int(outimc);

    /* Initialise the output map. This is because the output data array 
       probably still has the results of the nearest neighbour stack */

    for (i = 0; i < npts; i++) {
	outdata[i] = 0.0;
	outdatac[i] = 0;
    }

    /* Ok, loop for each of the input files and collect all the information
       for each pixel */

    for (i = 0; i < nim; i++) {
	dd = fileptrs + i;
	data = dd->data;
	cdata = dd->cdata;
	wt = dd->weight;
	nx = dd->nx;
	ny = dd->ny;

	/* Loop for each pixel in the data array and work out where it belongs
	   in the output array */

	n = 0;
	for (iy = 0; iy < ny; iy++) {
	    ind1 = iy*nx;
	    for (ix = 0; ix < nx; ix++) {
		ind = ind1 + ix;
		outloc(dd->vwcs,(double)ix,(double)iy,outwcs,dd->trans,
		       &oxf,&oyf);
 		if (oxf < 0.0 || oyf < 0.0)
 		    continue;
		ox = (int)oxf;
		oy = (int)oyf;
		if ((dd->bpm)[ind] != 0)
		    continue;
		if (ox < 0 || ox >= nxo || oy < 0 || oy >= nyo ||
		    cdata[ind] == 0) 
		    continue;
		dx = oxf - ox;
		dy = oyf - oy;
		w[0] = (1.0-dx)*(1.0-dy);
		w[1] = dx*(1.0-dy);
		w[2] = dy*(1.0-dx);
		w[3] = dx*dy;
		m = 0;
		value = data[ind]*(dd->expscale) + dd->skydiff;
		cval = (float)cdata[ind];
		for (jj = 0; jj <= 1; jj++) {
		    for (ii = 0; ii <= 1; ii++) {
			nn = (oy+jj)*nxo + ox + ii;
			if (nn >= npts || ox+ii >= nxo || oy+jj >= nyo)
			    continue;
			outdata[nn] += w[m]*value*wt*cval;
			workbufw[nn] += w[m]*wt*cval;
			m++;
		    }
		}
	    }
	}
    }

    /* Now do the final averaging */

    for (i = 0; i < npts; i++) {
	if (workbufw[i] != 0.0)
   	    outdata[i] /= workbufw[i];
	else
	    outdata[i] = fileptrs->sky;
 	workbufw[i] /= sumweight;
   }
    vircam_qmedsig(workbufw,NULL,npts,3.0,2,0.0,65535.0,&med,&noise);
    for (i = 0; i < npts; i++) {
	workbufw[i] *= (100.0/med);
	outdatac[i] = max(0,min(1000,vircam_nint(workbufw[i])));
    }
    freespace(workbufw);
    return(VIR_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        stack_nn
    \par Purpose:
        Do a stack with nearest neighbour interpolation
    \par Description:
        Do a stack with nearest neighbour interpolation. Rejection is done 
	here. Any pixels rejected will also be flagged in the input image
	masks so that this information can be used by other stacking
	routines.
    \par Language:
        C
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int stack_nn(void) {
    float *workbuf,*workbufc,*workbufw,lclip,hclip,*data,wt,med,noise,outvalc;
    float value,*dbuf,*wbuf,*cbuf,outval,avlev,avvar,*confwork;
    double oxf,oyf;
    int *iloc,i,npts,ix,iy,ind1,ind,ox,oy,nn,bufloc,ncontrib,*lbuf;
    int j,*cdata,nx,ny,n,nz;
    short int clipped_low,clipped_high;
    dstrct *dd;
    unsigned char *id,*clipmap,*cc,*cclast,*ibuf;
    short int *nbuf;
    cpl_image *outim,*outimc;

    /* Output file info */

    outim = vircam_fits_get_image(outfits);
    outimc = vircam_fits_get_image(outconf);
    nxo = (int)cpl_image_get_size_x(outim);
    nyo = (int)cpl_image_get_size_y(outim);

    /* Get workspace for collecting info over the whole output map */

    nz = nim + 2;
    workbuf = cpl_calloc(nz*nxo*nyo,sizeof(*workbuf));
    workbufc = cpl_calloc(nz*nxo*nyo,sizeof(*workbufc));
    workbufw = cpl_calloc(nz*nxo*nyo,sizeof(*workbufw));
    id = cpl_calloc(nz*nxo*nyo,sizeof(*id));
    iloc = cpl_calloc(nz*nxo*nyo,sizeof(*iloc));
    nbuf = cpl_calloc(nxo*nyo,sizeof(*nbuf));
    confwork = cpl_calloc(nxo*nyo,sizeof(*confwork));

    /* Workspace for marking output rows with clipped pixels */

    clipmap = cpl_calloc(2*nxo,sizeof(*clipmap));

    /* Buffers for averaging an individual pixel */

    dbuf = cpl_calloc(2*nz,sizeof(*dbuf));
    wbuf = cpl_calloc(2*nz,sizeof(*wbuf));
    cbuf = cpl_calloc(2*nz,sizeof(*cbuf));
    ibuf = cpl_calloc(2*nz,sizeof(*ibuf));
    lbuf = cpl_calloc(2*nz,sizeof(*lbuf));

    /* Get output buffers */

    outdata = cpl_image_get_data_float(outim);
    outdatac = cpl_image_get_data_int(outimc);

    /* Rejection thresholds */

    lclip = fileptrs->sky - lsig*(fileptrs->noise);
    hclip = fileptrs->sky + hsig*(fileptrs->noise);

    /* Ok, loop for each of the input files and collect all the information
       for each pixel */

    for (i = 0; i < nim; i++) {
	dd = fileptrs + i;
	nx = dd->nx;
	ny = dd->ny;
	npts = nx*ny;
	data = dd->data;
	cdata = dd->cdata;
	wt = dd->weight;

	/* Loop for each pixel in the data array and work out where it belongs
	   in the output array */

	n = 0;
	for (iy = 0; iy < ny; iy++) {
	    ind1 = iy*nx;
	    for (ix = 0; ix < nx; ix++) {
		ind = ind1 + ix;
		outloc(dd->vwcs,(double)ix,(double)iy,outwcs,dd->trans,
		       &oxf,&oyf);
		ox = vircam_nint(oxf) - 1;
		oy = vircam_nint(oyf) - 1;
		if (ox < 0 || ox >= nxo || oy < 0 || oy >= nyo || 
		    cdata[ind] == 0)
		    continue;
		nn = oy*nxo + ox;
		bufloc = oy*nxo*nz + ox*nz + nbuf[nn];
		value = data[ind]*(dd->expscale) + dd->skydiff;
		workbuf[bufloc] = value;
		workbufw[bufloc] = wt*(float)cdata[ind];
		workbufc[bufloc] = (float)cdata[ind];
		id[bufloc] = i;
		nbuf[nn] += 1;
		iloc[bufloc] = ind;
	    }
	}
    }


    /* Now loop through the output arrays and form an initial estimate */

    for (iy = 0; iy < nyo; iy++) {
	ind1 = iy*nxo;
	cc = clipmap + (iy % 2)*nxo;
	cclast = clipmap + ((iy-1) % 2)*nxo;
	for (ix = 0; ix < nxo; ix++) {
	    ind = ind1 + ix;
	    bufloc = iy*nxo*nz + ix*nz;
	    ncontrib = nbuf[ind];
	    if (ncontrib == 0) {
		outdata[ind] = fileptrs->sky;
		outdatac[ind] = 0;
		continue;
	    } 
	    cc[ix] = 0;

	    /* Put some stuff in buffers for the averaging routine */
	    
	    for (i = 0; i < ncontrib; i++) {
		dbuf[i] = workbuf[bufloc+i];
		wbuf[i] = workbufw[bufloc+i];
		cbuf[i] = workbufc[bufloc+i];
		ibuf[i] = id[bufloc+i];
		lbuf[i] = iloc[bufloc+i];
	    }

	    /* Do the averages with the nominal clipping */

            do_averages(ncontrib,dbuf,wbuf,cbuf,ibuf,lclip,hclip,0.0,
			&clipped_low,&clipped_high,&outval,&outvalc);
	    
	    /* Store these away */

	    outdata[ind] = outval;
	    confwork[ind] = outvalc;
	    cc[ix] = (clipped_high >= 0);
	    if (clipped_low >= 0) 
		((fileptrs+ibuf[clipped_low])->bpm)[lbuf[clipped_low]] = 2;
	    if (clipped_high >= 0) 
		((fileptrs+ibuf[clipped_high])->bpm)[lbuf[clipped_high]] = 1;
	}

	/* Look at the non-edge pixels and see if the local variation can
	   show whether any clipping that has been done is justified. NB:
	   we are looking at the _previous_ row here */

	if (iy < 2)
	    continue;
	for (ix = 1; ix < nxo-1; ix++) {
	    if (! cclast[ix])
		continue;
	    ind = (iy-1)*nxo + ix;
	    bufloc = (iy-1)*nxo*nz + ix*nz;
	    ncontrib = nbuf[ind];
	    if (ncontrib == 0)
		continue;
	    avlev = 0.0;
	    for (i = -1; i <= 1; i++) 
		for (j = -1; j <= 1; j++) 
		    avlev += outdata[ind + i*nxo + j];
	    avlev /= 9.0;
	    avvar = 0.0;
	    for (i = -1; i <= 1; i++) 
		for (j = -1; j <= 1; j++) 
		    avvar += (float)fabs(avlev - outdata[ind + i*nxo + j]);
	    avvar /= 9.0;
	    if (avlev <= hclip && avvar <= 2.0*(fileptrs->sky))
		continue;

	    /* Put some stuff in buffers for the averaging routine */
	    
	    for (i = 0; i < ncontrib; i++) {
		dbuf[i] = workbuf[bufloc+i];
		wbuf[i] = workbufw[bufloc+i];
		cbuf[i] = workbufc[bufloc+i];
		ibuf[i] = id[bufloc+i];
		lbuf[i] = iloc[bufloc+i];
		(fileptrs+ibuf[i])->bpm[lbuf[i]] = 0;
	    }

	    /* Redo the averages with the nominal clipping */

            do_averages(ncontrib,dbuf,wbuf,cbuf,ibuf,lclip,hclip,avvar,
			&clipped_low,&clipped_high,&outval,&outvalc);

	    /* Store these away */

	    outdata[ind] = outval;
	    confwork[ind] = outvalc;
	    if (clipped_low >= 0) 
		((fileptrs+ibuf[clipped_low])->bpm)[lbuf[clipped_low]] = 2;
	    if (clipped_high >= 0) 
		((fileptrs+ibuf[clipped_high])->bpm)[lbuf[clipped_high]] = 1;
	}    
    }

    /* Now renormalise the confidence map */

    npts = nxo*nyo;
    vircam_qmedsig(confwork,NULL,npts,3.0,2,25.0,65535.0,&med,&noise);
    for (i = 0; i < npts; i++) {
	confwork[i] *= (100.0/med);
	outdatac[i] = max(0,min(1000,vircam_nint(confwork[i])));
    }

    /* Free up some workspace */

    freespace(workbuf);
    freespace(workbufc);
    freespace(workbufw);
    freespace(confwork);
    freespace(id);
    freespace(iloc);
    freespace(nbuf);
    freespace(clipmap);
    freespace(dbuf);
    freespace(wbuf);
    freespace(cbuf);
    freespace(ibuf);
    freespace(lbuf);
    return(VIR_OK);
}
		    
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        do_averages
    \par Purpose:
        Work out the average for a given pixel.
    \par Description:
        Work out the average for a given pixel. Keep track of which pixels
	have been clipped.
    \par Language:
        C
    \param ncontrib
        The number of contributing pixels
    \param data
        The data array of input pixels
    \param wconf
        The data array of weighted confidence
    \param conf
        The data array of confidence
    \param id
        The data array of frame ids
    \param lclip
        The lower clipping level
    \param hclip
        The upper clipping level
    \param extra
        An extra upper clipping level
    \param lowcl
        The location in the data arrays of a low clipped pixel
    \param highcl
        The location in the data arrays of a high clipped pixel
    \param outval
        The output value for the pixel
    \param outvalc
        The output confidence value
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/
    
static void do_averages(int ncontrib, float *data, float *wconf, float *conf,
			unsigned char *id, float lclip, float hclip, 
			float extra, short int *lowcl, short int *highcl, 
			float *outval, float *outvalc) {
    float minval,maxval,sum,wsum,csum,reflev,clipval;
    int imax,imin,i;

    /* Do the summation and keep track of the minimum and
       maximum values */

    minval = 1.0e10;
    maxval = -1.0e10;
    sum = 0.0;
    wsum = 0.0;
    csum = 0.0;
    imin = -1;
    imax = -1;
    for (i = 0; i < ncontrib; i++) {
	sum += data[i]*wconf[i];
	wsum += wconf[i];
	csum += conf[i];
	if (data[i] > maxval) {
	    maxval = data[i];
	    imax = i;
	}
	if (data[i] < minval) {
	    minval = data[i];
	    imin = i;
	}
    }

    /* First crack at output value */

    if (wsum > 0.0) {
	*outval = sum/wsum;
    } else {
	*outval = fileptrs->sky;
    }

    /* Look at high clipping for of cosmic rays */

    *highcl = -1;
    if (maxval > hclip && wsum > 150.0 && csum > 150.0 && 
	imin != -1 && imax != -1) {
	reflev = (sum - data[imax]*wconf[imax])/(wsum - wconf[imax]);
	clipval = reflev + hsig*(fileptrs+id[imax])->noise;
	clipval *= max(1.0,reflev)/max(1.0,(fileptrs+id[imax])->sky);
	clipval += extra;

	/* Clip the maximum point */

	if (maxval > clipval) {
	    sum -= data[imax]*wconf[imax];
	    wsum -= wconf[imax];
	    *outval = reflev;
	    *highcl = imax;
	}
    }

    /* Clip low points (presumably serious cosmetic issues on detectors) */

    *lowcl = -1;
    if (minval < lclip  && wsum > 150.0 && csum > 150.0 &&
	imin != -1 && imax != -1) {
	reflev = (sum - data[imin]*wconf[imin])/(wsum - wconf[imin]);
	clipval = reflev + hsig*(fileptrs+id[imin])->noise;
	clipval *= max(1.0,reflev)/max(1.0,(fileptrs+id[imin])->sky);
	if (minval < clipval) {
	    sum -= data[imin]*wconf[imin];
	    wsum -= wconf[imin];
	    *outval = reflev;
	    *lowcl = imin;
	}
    }

    /* Output confidence (not normalised) */

    *outvalc = wsum/sumweight;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        output_files
    \par Purpose:
        Set up the output files
    \par Description:
        Work out the size of the output files given the coverage of each
	of the input files. Set up the output files
    \par Language:
        C
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void output_files(void) {
    cpl_image *newim,*newconf;
    double lowerleft[2],upperleft[2],lowerright[2],upperright[2],xout,yout;
    double xmin,ymin,sh[2];
    dstrct *dd;
    int i,ixo,iyo;

    /* Initialise the first element of the results arrays since we're 
       doing all this relative to the first image */

    lowerleft[0] = 1.0;
    lowerleft[1] = 1.0;
    upperleft[0] = 1.0;
    upperleft[1] = (double)fileptrs->ny;
    lowerright[0] = (double)fileptrs->nx;
    lowerright[1] = 1.0;
    upperright[0] = (double)fileptrs->nx;
    upperright[1] = (double)fileptrs->ny;

    /* Loop for all images relative to the first one */

    for (i = 1; i < nim; i++) {
	dd = fileptrs + i;
	outloc(dd->vwcs,(double)1.0,(double)1.0,fileptrs->vwcs,dd->trans,
	       &xout,&yout);
	lowerleft[0] = min(lowerleft[0],xout);
	lowerleft[1] = min(lowerleft[1],yout);
	outloc(dd->vwcs,(double)1.0,(double)dd->ny,fileptrs->vwcs,dd->trans,
	       &xout,&yout);
	upperleft[0] = min(upperleft[0],xout);
	upperleft[1] = max(upperleft[1],yout);
	outloc(dd->vwcs,(double)dd->nx,(double)1.0,fileptrs->vwcs,dd->trans,
	       &xout,&yout);
	lowerright[0] = max(lowerright[0],xout);
	lowerright[1] = min(lowerright[1],yout);
	outloc(dd->vwcs,(double)dd->nx,(double)dd->ny,fileptrs->vwcs,dd->trans,
	       &xout,&yout);
	upperright[0] = max(upperright[0],xout);
	upperright[1] = max(upperright[1],yout);
    }

    /* Ok, what are the limits? */

    ixo = vircam_nint(max(lowerright[0]-lowerleft[0],upperright[0]-upperleft[0])) + 1;
    iyo = vircam_nint(max(upperright[1]-lowerright[1],upperleft[1]-lowerleft[1])) + 1;
    xmin = min(lowerleft[0],upperleft[0]);
    ymin = min(lowerleft[1],lowerright[1]);
    
    /* Create the image */

    newim = cpl_image_new((cpl_size)ixo,(cpl_size)iyo,CPL_TYPE_FLOAT);
    outfits = vircam_fits_wrap(newim,fileptrs->fname,NULL,NULL);
    cpl_propertylist_update_int(vircam_fits_get_ehu(outfits),"NAXIS1",ixo);
    cpl_propertylist_update_int(vircam_fits_get_ehu(outfits),"NAXIS2",iyo);

    /* Update the reference point for the WCS */

    sh[0] = xmin;
    sh[1] = ymin;
    (void)vircam_crpixshift(vircam_fits_get_ehu(outfits),1.0,sh);
    outwcs = getwcsinfo(vircam_fits_get_ehu(outfits));

    /* Now create the confidence map image */

    newconf = cpl_image_new((cpl_size)ixo,(cpl_size)iyo,CPL_TYPE_INT);
    outconf = vircam_fits_wrap(newconf,outfits,NULL,NULL);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        diffxy
    \par Purpose:
        Work out the cartesian offset between two images
    \par Description:
        Work out the cartesian offset between two images by using the
	WCS information read from the input headers. The result is in the
	sense of reference - programme.
    \par Language:
        C
    \param wref
        The WCS descriptor for the reference image
    \param wprog
        The WCS descriptor for the programme image
    \param dx
        The output X offset
    \param dy
        The output Y offset
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void diffxy(mywcs *wref, mywcs *wprog, float *dx, float *dy) {
    double xref,yref,xprog,yprog;

    /* Find the middle of the reference frame */

    xref = 0.5*(double)wref->nx;
    yref = 0.5*(double)wref->ny;

    /* Work out the output position */

    outloc(wref,xref,yref,wprog,NULL,&xprog,&yprog);

    /* Now the offsets */

    *dx = (float)(xref - xprog);
    *dy = (float)(yref - yprog);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        outloc
    \par Purpose:
        Work the output location of an input pixel
    \par Description:
        Given an input pixel position and input and output WCS descriptors,
	work out the output pixel position. If a coordinate transformation
	refinement has been done, the apply it here.
    \par Language:
        C
    \param win
        The input WCS descriptor
    \param xin
        The input X position
    \param yin
        The input Y position
    \param wout
        The output WCS descriptor
    \param trans
        The coordinate transformation refinement coefficients
    \param xout
        The output X position
    \param yout
        The output Y position
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void outloc(mywcs *win, double xin, double yin, mywcs *wout, 
		   cpl_array *trans, double *xout, double *yout) {
    double xt,yt,xi,eta,r,rfac,aa,tandec,denom,rp,*tdata;

    /* Do the conversion. First to standard coordinates in the frame of
       the input image */

    xt = xin - win->crpix[0];
    yt = yin - win->crpix[1];
    xi = win->cd[0]*xt + win->cd[1]*yt;
    eta = win->cd[2]*xt + win->cd[3]*yt;
    r = sqrt(xi*xi + eta*eta);
    rfac = win->pv21 + win->pv23*pow(r,2.0) + win->pv25*pow(r,4.0);
    r /= rfac;
    rfac = win->pv21 + win->pv23*pow(r,2.0) + win->pv25*pow(r,4.0);
    xi /= rfac;
    eta /= rfac;
    aa = atan(xi*win->secd/(1.0-eta*win->tand));
    if (xi != 0.0) 
	tandec = (eta+win->tand)*sin(aa)/(xi*win->secd);
    else
	tandec = (eta+win->tand)/(1.0 - eta*win->tand);

    /* Now from standard coordinates in the frame of the output image */

    aa += (win->crval[0] - wout->crval[0]);
    denom = wout->tand*tandec + cos(aa);
    xi = wout->secd*sin(aa)/denom;
    eta = (tandec - wout->tand*cos(aa))/denom;
    rp = sqrt(xi*xi + eta*eta);
    rfac = wout->pv21 + wout->pv23*pow(rp,2.0) + wout->pv25*pow(rp,4.0);
    xi *= rfac;
    eta *= rfac; 
    denom = wout->cd[0]*wout->cd[3] - wout->cd[1]*wout->cd[2];
    xt = (xi*wout->cd[3] - eta*wout->cd[1])/denom + wout->crpix[0];
    yt = (eta*wout->cd[0] - xi*wout->cd[2])/denom + wout->crpix[1];
    
    /* If a fine tune transformation is defined, then do that now */

    if (trans != NULL) {
	tdata = cpl_array_get_data_double(trans);
	*xout = xt*tdata[0] + yt*tdata[1] + tdata[2];
	*yout = xt*tdata[3] + yt*tdata[4] + tdata[5];
    } else {
	*xout = xt;
	*yout = yt;
    }
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        backgrnd_ov
    \par Purpose:
        Work out the background correction between two images
    \par Description:
        Work out the background correction between two images by examining
	the region of common coverage between them. If this common region is
	less than half of an image, then the whole of each image to work out
	a background estimate.
    \par Language:
        C
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void backgrnd_ov(void) {
    int i,dx,dy,*cdata,n,jj,ind1,nx,ii,ind;
    long npts1,npts,start[2],end[2];
    dstrct *dd1,*dd2;
    float *data,sky1,skynoise1,sky2,skynoise2,frac;

    /* Initialise a few things */

    dd1 = fileptrs;
    npts1 = dd1->nx*dd1->ny;

    /* Start by working out the full background of the first frame */

    skyest(dd1->data,dd1->cdata,npts1,3.0,&sky1,&skynoise1);
    dd1->sky = sky1;
    dd1->noise = skynoise1;
    dd1->sky /= dd1->expscale;
    dd1->skydiff = 0.0;
    dd1->weight = 1.0;
    sumweight = 1.0;

    /* Now loop for all the others */    

    for (i = 1; i < nim; i++) {
	dd2 = fileptrs + i;

	/* Offset differences */

	dx = vircam_nint(dd2->xoff - dd1->xoff);
	dy = vircam_nint(dd2->yoff - dd1->yoff);

	/* Ok, here's the tricky bit. Work out the range of x and y for
	   each of the images. The shift moves an image to the left and
	   and down. Hence positive values of dx or dy moves the second
	   frame further to the left of down compared to the first frame.
	   Here are the coordinate ranges for the first image */

	start[0] = max(0,-dx);
	start[1] = max(0,-dy);
	end[0] = min(dd1->nx-1,(dd1->nx)-dx-1);
	end[1] = min(dd1->ny-1,(dd1->ny)-dy-1);
	nx = dd1->nx;
	
	/* How much does this cover the first image? */

	npts = (end[0]-start[0]+1)*(end[1]-start[1]+1);
	frac = (float)npts/(float)npts1;

	/* If the coverage is more than 50% then calculate the background
	   in the coverage region for the first image */

	if (frac >= 0.5) {
 	    data = cpl_malloc(npts*sizeof(*data));
 	    cdata = cpl_malloc(npts*sizeof(*cdata));
	    n = 0;
	    for (jj = start[1]; jj <= end[1]; jj++) {
		ind1 = jj*nx;
		for (ii = start[0]; ii <= end[0]; ii++) {
		    ind = ind1 + ii;
		    data[n] = (dd1->data)[ind];
		    cdata[n++] = (dd1->cdata)[ind];
		}
	    }
	    skyest(data,cdata,npts,3.0,&sky1,&skynoise1);
            freespace(data);
	    freespace(cdata);
	} else {
	    sky1 = dd1->sky;
	}

	/* And here are the coordinate ranges for the second image if 
	   the coverage is more than 50%. Get the subset you need. */

	if (frac > 0.5) {
  	    start[0] = max(1,dx);
 	    start[1] = max(1,dy);
	    end[0] = min(dd2->nx-1,(dd2->nx)+dx-1);
	    end[1] = min(dd2->ny-1,(dd2->ny)+dy-1);
	    nx = dd2->nx;
 	    npts = (end[0]-start[0]+1)*(end[1]-start[1]+1);
 	    data = cpl_malloc(npts*sizeof(*data));
	    cdata = cpl_malloc(npts*sizeof(*cdata));
	    n = 0;
	    for (jj = start[1]; jj <= end[1]; jj++) {
		ind1 = jj*nx;
		for (ii = start[0]; ii <= end[0]; ii++) {
		    ind = ind1 + ii;
		    data[n] = dd2->data[ind];
		    cdata[n++] = dd2->cdata[ind];
		}
	    }
	    skyest(data,cdata,npts,3.0,&sky2,&skynoise2);
	    freespace(data);
	    freespace(cdata);
	} else {		
	    npts = (dd2->nx)*(dd2->ny);
	    skyest(dd2->data,dd2->cdata,npts,3.0,&sky2,&skynoise2);
	}

	/* Store info away */

	dd2->sky = sky2/dd2->expscale;
	dd2->skydiff = sky1 - sky2;
	dd2->noise = skynoise2;
	if (dd2->noise != 0.0 && dd2->sky != 0.0) 
  	    dd2->weight = (float)(pow(dd1->noise,2.0)/pow(dd2->noise,2.0)); 
	else 
	    dd2->weight = 0.0;
	sumweight += dd2->weight;
    }
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        skyest
    \par Purpose:
        Work out a sky estimate
    \par Description:
        Work out a sky estimate given a data array and confidence map.
	Clipping is done at a given threshold.
    \par Language:
        C
    \param data
        The input data array
    \param cdata
        The input confidence map array
    \param npts
        The number of entries in the data and confidence arrays
    \param thresh
        The clipping threshold in units of background noise
    \param skymed
        The background median
    \param skynoise
        The background noise
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void skyest(float *data, int *cdata, long npts, float thresh, 
		   float *skymed, float *skynoise) {
    unsigned char *bpm;
    int i;

    /* Get a dummy bad pixel mask */

    bpm = cpl_calloc(npts,sizeof(*bpm));
    for (i = 0; i < npts; i++)
	bpm[i] = (cdata[i] == 0);

    /* Get the stats */

    vircam_qmedsig(data,bpm,npts,thresh,2,-1000.0,65535.0,skymed,skynoise);

    /* Clean up */

    freespace(bpm);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        mknewtab
    \par Purpose:
        Create a new catalogue table with update x,y coordinates
    \par Description:
        Create a new catalogue table. The x,y coordinates are updated
	to a new coordinate system defined by an input wcs descriptor.
    \par Language:
        C
    \param catref
        The input catalogue of the reference image
    \param dd
        The file structure for the frame to which we want to transform the
	coordinates
    \returns 
        A copy of the input catalogue, but with the x,y coordinates now 
	in the reference frame of another image
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static cpl_table *mknewtab(vir_tfits *catref, dstrct *dd) {
    cpl_table *newtab;
    double xx,yy,xx2,yy2;
    int nr,i;
    float *x,*y;

    /* Make a copy of the input table */

    newtab = cpl_table_duplicate(vircam_tfits_get_table(catref));
    nr = (int)cpl_table_get_nrow(newtab);

    /* Get the columns of interest */

    x = cpl_table_get_data_float(newtab,"X_coordinate");
    y = cpl_table_get_data_float(newtab,"Y_coordinate");

    /* Loop for each object */

    for (i = 0; i < nr; i++) {
	xx = (double)x[i];
	yy = (double)y[i];
	outloc(fileptrs->vwcs,xx,yy,dd->vwcs,NULL,&xx2,&yy2);
	x[i] = (float)xx2;
	y[i] = (float)yy2;
    }
    return(newtab);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        transinit
    \par Purpose:
        Initialise transformation array
    \par Description:
        Initialise the transformation array
    \par Language:
        C
    \returns 
        A transformation array with no offsets, unit scale and no rotation.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/
	    
static cpl_array *transinit(void) {
    double *td;
    cpl_array *t;

    t = cpl_array_new(6,CPL_TYPE_DOUBLE);
    td = cpl_array_get_data_double(t);
    td[0] = 1.0;
    td[1] = 0.0;
    td[2] = 0.0;
    td[3] = 0.0;
    td[4] = 1.0;
    td[5] = 0.0;
    return(t);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        fileptrs_close
    \par Purpose:
        Clean up and free fileptrs array
    \par Description:
        Clean up and free fileptrs array
    \par Language:
        C
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void fileptrs_close(void) {
    int i;

    for (i = 0; i < nim; i++) {
        freespace((fileptrs+i)->vwcs);
	freearray((fileptrs+i)->trans);
    }
    freespace(fileptrs);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        getwcsinfo
    \par Purpose:
        Extract WCS info from a header
    \par Description:
        Extract WCS info from a header. Angles are all converted to 
	radians. Calculate a few convenience values
    \par Language:
        C
    \returns
        A WCS descriptor
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static mywcs *getwcsinfo(cpl_propertylist *plist) {
    mywcs *w;
    int i;

    /* Copy over the relevant info */

    w = cpl_malloc(sizeof(mywcs));
    (void)vircam_pfits_get_crval1(plist,&(w->crval[0]));
    (void)vircam_pfits_get_crval2(plist,&(w->crval[1]));
    (void)vircam_pfits_get_crpix1(plist,&(w->crpix[0]));
    (void)vircam_pfits_get_crpix2(plist,&(w->crpix[1]));
    (void)vircam_pfits_get_cd11(plist,&(w->cd[0]));
    (void)vircam_pfits_get_cd12(plist,&(w->cd[1]));
    (void)vircam_pfits_get_cd21(plist,&(w->cd[2]));
    (void)vircam_pfits_get_cd22(plist,&(w->cd[3]));
    (void)vircam_pfits_get_pv21(plist,&(w->pv21));
    (void)vircam_pfits_get_pv23(plist,&(w->pv23));
    (void)vircam_pfits_get_pv25(plist,&(w->pv25));
    (void)vircam_pfits_get_naxis1(plist,&(w->nx));
    (void)vircam_pfits_get_naxis2(plist,&(w->ny));

    /* Change angles to radians */

    for (i = 0; i < 2; i++) 
	w->crval[i] /= DEGRAD;
    for (i = 0; i < 4; i++)
	w->cd[i] /= DEGRAD;

    /* Add a couple of convenience values */

    w->tand = tan(w->crval[1]);
    w->secd = 1.0/cos(w->crval[1]);

    /* Get out of here */

    return(w);
}

static void tidy(void) {
    fileptrs_close();
    freespace(outwcs);
}

/* 

$Log: vircam_imstack.c,v $
Revision 1.15  2012/01/15 17:40:09  jim
Minor modifications to take into accout the changes in cpl API for v6

Revision 1.14  2010/09/09 12:11:09  jim
Fixed problems with docs that make doxygen barf

Revision 1.13  2010/06/03 12:15:31  jim
A few mods to get rid of compiler warnings

Revision 1.12  2009/06/25 14:14:50  jim
Fixed a little error handling problem

Revision 1.11  2009/06/01 04:57:19  jim
tidied up a few things

Revision 1.10  2009/05/31 05:54:55  jim
Nothing

Revision 1.9  2009/05/31 05:42:55  jim
reweighting linear interpolation

Revision 1.8  2009/05/30 20:38:41  jim
fixed another normalisation error

Revision 1.7  2009/05/30 19:41:04  jim
Fixed another small bug in the confidence map scaling

Revision 1.6  2009/05/29 15:39:49  jim
changed the way confidence map is normalised

Revision 1.5  2009/02/20 10:52:49  jim
Removed some superfluous declarations and tidied up some things to get
rid of compiler warnings

Revision 1.4  2009/01/28 10:55:10  jim
Fixed so that if the calculation of trans array fails, then it issues a
warning, but keeps going

Revision 1.3  2009/01/19 14:35:43  jim
Modified search radii in matching routines

Revision 1.2  2008/11/25 06:19:20  jim
Added routine prologues

Revision 1.1  2008/11/21 10:19:16  jim
New routine


*/
