/* $Id: vircam_nditcor.c,v 1.4 2009/05/20 12:18:42 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2005 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2009/05/20 12:18:42 $
 * $Revision: 1.4 $
 * $Name: vcam-1_3_0 $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cpl.h>
#include "vircam_mods.h"
#include "vircam_utils.h"
#include "vircam_fits.h"

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Correct input data for dark current
  
    \par Name:
        vircam_nditcor
    \par Purpose:
        Correct input data for multiple dits.
    \par Description:
        Divide the input data by the number of dits. This is so that
	all images are scaled the same way. The result overwrites
	the input data array.
    \par Language:
        C
    \param infile 
        The input data image (overwritten by result).
    \param ndit
        The number of dits in the observation
    \param status 
        An input/output status that is the same as the returned values below.
    \retval VIR_OK 
        if everything is ok
    \retval VIR_FATAL 
        if image data fails to load
    \par QC headers:
        None
    \par DRS headers:
        The following DRS keywords are written to the infile extension header
	- \b NDITCOR
	    A logical keyword to say that this correction has been applied
    \par Main headers:
        The following keywords in the non-hierarchical part of the headers
	have been modified:
	- \b EXPTIME
	    The exposure time on entry is divided by the number of DITs. This
	    is done both in the extension and the primary header.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_nditcor(vir_fits *infile, int ndit, int *status) {
    cpl_image *im;
    double dnd,exptime;
    cpl_propertylist *oplist;
    char comment[32];

    /* Inherited status */

    if (*status != VIR_OK)
	return(*status);

    /* Do we need to be here? */

    if (cpl_propertylist_has(vircam_fits_get_ehu(infile),"ESO DRS NDITCOR"))
	return(*status);

    /* Get the image and the scale factor */
    
    im = vircam_fits_get_image(infile);
    dnd = (float)ndit;
    
    /* Divide it through... */

    cpl_image_divide_scalar(im,dnd);

    /* Now put some stuff in the DRS extension and update the exposure
       time to reflect this correction. The latter has to be done in
       both the extension and primary headers */

    oplist = vircam_fits_get_ehu(infile);
    if (oplist != NULL) {
	cpl_propertylist_update_bool(oplist,"ESO DRS NDITCOR",1);
	(void)sprintf(comment,"Corrected for ndit=%d",ndit);
	cpl_propertylist_set_comment(oplist,"ESO DRS NDITCOR",comment);
	exptime = cpl_propertylist_get_double(oplist,"EXPTIME");
	exptime /= dnd;
	cpl_propertylist_update_double(oplist,"EXPTIME",exptime);
	cpl_propertylist_set_comment(oplist,"EXPTIME",comment);
    } 
    oplist = vircam_fits_get_phu(infile);
    if (oplist != NULL) {
	(void)sprintf(comment,"Corrected for ndit=%d",ndit);
	exptime = cpl_propertylist_get_double(oplist,"EXPTIME");
	exptime /= dnd;
	cpl_propertylist_update_double(oplist,"EXPTIME",exptime);
	cpl_propertylist_set_comment(oplist,"EXPTIME",comment);
    } 

    /* Get out of here */

    GOOD_STATUS
}


/**@}*/

/*

$Log: vircam_nditcor.c,v $
Revision 1.4  2009/05/20 12:18:42  jim
Modified so that if the operation is already done, then it just returns

Revision 1.3  2009/02/20 10:55:20  jim
removed unnecessary declaration

Revision 1.2  2009/01/20 09:31:37  jim
Fixed typo

Revision 1.1  2009/01/19 14:32:57  jim
New routine


*/
